/*
 * @OSF_FREE_FREE_COPYRIGHT@
 * 
 */
/*
 * HISTORY
 * $Log: signal.c,v $
 * Revision 1.1.2.9  1997/02/07  14:33:12  barbou
 * 	Use flush_instruction_cache_range() to flush the signal trampoline.
 * 	[1997/02/07  13:45:24  barbou]
 *
 * Revision 1.1.2.8  1996/12/03  08:25:49  stephen
 * 	Synchronize instruction/data caches after writing
 * 	trampoline code into stack frame
 * 	[1996/12/03  08:25:35  stephen]
 * 
 * Revision 1.1.2.7  1996/11/20  16:51:28  barbou
 * 	Fix for the "Stacked signals" bug: account for a stack frame
 * 	when setting up the stack pointer in sys_sigreturn.
 * 	Also disabled the "Stacked signals" message.
 * 	[1996/11/20  16:50:50  barbou]
 * 
 * Revision 1.1.2.6  1996/11/05  15:38:40  barbou
 * 	Adapted to the split pt_regs structure.
 * 	[96/11/05            barbou]
 * 
 * Revision 1.1.2.5  1996/10/03  15:22:42  barbou
 * 	Fixed a typo.
 * 	[96/10/03            barbou]
 * 
 * Revision 1.1.2.4  1996/09/27  22:30:47  barbou
 * 	Fixed typo.
 * 	[1996/09/27  22:15:20  barbou]
 * 
 * 	Merged up to linux-2.0.21.
 * 	[1996/09/27  22:01:36  barbou]
 * 
 * Revision 1.1.2.3  1996/09/26  15:59:43  barbou
 * 	Check for "regs->trap == 0x0c00" instead of "regs->orig_gpr3 >= 0" to
 * 	check if called from a system call.
 * 	[96/09/26            barbou]
 * 
 * Revision 1.1.2.2  1996/09/19  14:39:30  barbou
 * 	Changed OSFMACH3 conditionals to CONFIG_OSFMACH3 and
 * 	make sure that the configuration header file is included.
 * 	[1996/09/19  14:32:43  barbou]
 * 
 * Revision 1.1.2.1  1996/09/18  14:20:14  barbou
 * 	Created.
 * 	[1996/09/16  13:55:31  barbou]
 * 
 * $EndLog$
 */

/*
 *  linux/arch/ppc/kernel/signal.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *  Adapted for PowerPC by Gary Thomas
 */

#include <linux/autoconf.h>

#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/kernel.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/wait.h>
#include <linux/ptrace.h>
#include <linux/unistd.h>
#include <asm/segment.h>

#ifdef	CONFIG_OSFMACH3
#include <osfmach3/server_thread.h>

#include <asm/pgtable.h>
#endif	/* CONFIG_OSFMACH3 */

#define _S(nr) (1<<((nr)-1))

#define _BLOCKABLE (~(_S(SIGKILL) | _S(SIGSTOP)))

#define DEBUG_SIGNALS
#undef  DEBUG_SIGNALS

#define PAUSE_AFTER_SIGNAL
#undef  PAUSE_AFTER_SIGNAL

asmlinkage int sys_waitpid(pid_t pid,unsigned long * stat_addr, int options);
#ifdef	CONFIG_OSFMACH3
asmlinkage int do_signal(unsigned long oldmask, struct pt_regs * regs);
#endif	/* CONFIG_OSFMACH3 */

/*
 * atomically swap in the new signal mask, and wait for a signal.
 */
asmlinkage int sys_sigsuspend(unsigned long set, int p2, int p3, int p4, int p6, int p7, struct pt_regs *regs)
{
	unsigned long mask;

	mask = current->blocked;
	current->blocked = set & _BLOCKABLE;
	regs->gpr[3] = -EINTR;
#ifdef DEBUG_SIGNALS
#ifdef	CONFIG_OSFMACH3
printk("Task: %p[%d] - SIGSUSPEND at %lx, Mask: %lx\n", current, current->pid, regs->nip, set);	
#else	/* CONFIG_OSFMACH3 */
printk("Task: %x[%d] - SIGSUSPEND at %x, Mask: %x\n", current, current->pid, regs->nip, set);	
#endif	/* CONFIG_OSFMACH3 */
#endif
	while (1) {
		current->state = TASK_INTERRUPTIBLE;
		schedule();
		if (do_signal(mask,regs))
			/*
			 * If a signal handler needs to be called,
			 * do_signal() has set R3 to the signal number (the
			 * first argument of the signal handler), so don't
			 * overwrite that with EINTR !
			 * In the other cases, do_signal() doesn't touch 
			 * R3, so it's still set to -EINTR (see above).
			 */
			return regs->gpr[3];
	}
}

/*
 * This sets regs->esp even though we don't actually use sigstacks yet..
 */
asmlinkage int sys_sigreturn(struct pt_regs *regs)
{
	struct sigcontext_struct *sc;
	struct pt_regs *int_regs;
	int signo;
#ifdef	CONFIG_OSFMACH3
	struct pt_regs *extra_regs = (struct pt_regs *) regs->pt_extra_regs;
#endif	/* CONFIG_OSFMACH3 */

	sc = (struct sigcontext_struct *)(regs->gpr[1]+STACK_FRAME_OVERHEAD);
	current->blocked = get_fs_long(&sc->oldmask) & _BLOCKABLE;
	int_regs = (struct pt_regs *) get_fs_long(&sc->regs);
	signo = get_fs_long(&sc->signal);
	sc++;  /* Pop signal 'context' */
#ifdef DEBUG_SIGNALS
#ifdef	CONFIG_OSFMACH3
printk("Sig return - Regs: %p, sc: %p, sig: %d\n", int_regs, sc, signo);
#else	/* CONFIG_OSFMACH3 */
printk("Sig return - Regs: %x, sc: %x, sig: %d\n", int_regs, sc, signo);
#endif	/* CONFIG_OSFMACH3 */
#endif
	if (sc == (struct sigcontext_struct *)(int_regs))
	{ /* Last stacked signal */
#ifdef	CONFIG_OSFMACH3
		/*
		 * Copy the regular registers (common to Mach and Linux).
		 */
		memcpy_fromfs(regs, int_regs,
			      (((int)(&regs->pt_extra_regs)) -
			       ((int)(regs))));
		/*
		 * Copy the "extra registers" (which exist only in Linux).
		 */
		memcpy_fromfs(&extra_regs->pt_extra_regs + 1,
			      &int_regs->pt_extra_regs + 1,
			      (sizeof (*regs) -
			       (((int)(&regs->pt_extra_regs+1)) -
				((int)(regs)))));
		if (extra_regs->trap == 0x0c00 &&
		    ((int)extra_regs->result == -ERESTARTNOHAND ||
		     (int)extra_regs->result == -ERESTARTSYS ||
		     (int)extra_regs->result == -ERESTARTNOINTR))
		{
			regs->gpr[3] = extra_regs->orig_gpr3;
			regs->nip -= 4; /* Back up & retry system call */
			extra_regs->result = 0;
		}
#else	/* CONFIG_OSFMACH3 */
		memcpy(regs, int_regs, sizeof(*regs));
		if ((int)regs->orig_gpr3 >= 0 &&
		    ((int)regs->result == -ERESTARTNOHAND ||
		     (int)regs->result == -ERESTARTSYS ||
		     (int)regs->result == -ERESTARTNOINTR))
		{
			regs->gpr[3] = regs->orig_gpr3;
			regs->nip -= 4; /* Back up & retry system call */
			regs->result = 0;
		}
#endif	/* CONFIG_OSFAMCH3 */
#ifdef PAUSE_AFTER_SIGNALS
cnpause();
#endif
#ifdef	CONFIG_OSFMACH3
		return (extra_regs->result);
#else	/* CONFIG_OSFMACH3 */
		return (regs->result);
#endif	/* CONFIG_OSFMACH3 */
	} else
	{ /* More signals to go */
#ifdef DEBUG_SIGNALS
printk("More signals?\n");
#endif
#ifndef	CONFIG_OSFMACH3
printk("Stacked signals!\n");
#endif	/* CONFIG_OSFMACH3 */
		regs->gpr[1] = (unsigned long)sc - STACK_FRAME_OVERHEAD;
		regs->gpr[3] = get_fs_long(&sc->signal);
#ifdef	CONFIG_OSFMACH3
		int_regs = (struct pt_regs *) get_fs_long((long *) &sc->regs);
#else	/* CONFIG_OSFMACH3 */
		int_regs = (struct pt_regs *) get_fs_long(&sc->regs);
#endif	/* CONFIG_OSFMACH3 */
		regs->gpr[4] = (unsigned long)int_regs;
		regs->link = (unsigned long)(int_regs+1);
		regs->nip = get_fs_long(&sc->handler);
#ifdef	CONFIG_OSFMACH3
		return regs->gpr[3];
#else	/* CONFIG_OSFMACH3 */
		return (sc->signal);
#endif	/* CONFIG_OSFMACH3 */
	}
}


/*
 * Note that 'init' is a special process: it doesn't get signals it doesn't
 * want to handle. Thus you cannot kill init even with a SIGKILL even by
 * mistake.
 *
 * Note that we go through the signals twice: once to check the signals that
 * the kernel can handle, and then we build all the user-level signal handling
 * stack-frames in one go after that.
 */
asmlinkage int do_signal(unsigned long oldmask, struct pt_regs * regs)
{
	unsigned long mask = ~current->blocked;
	unsigned long handler_signal = 0;
	unsigned long *frame = NULL;
	unsigned long *trampoline;
	unsigned long *regs_ptr;
	unsigned long nip = 0;
	unsigned long signr;
	int bitno;
	struct sigcontext_struct *sc;
	struct sigaction * sa;
#ifdef	CONFIG_OSFMACH3
	unsigned long frame_size;
	struct pt_regs	*extra_regs = (struct pt_regs *) regs->pt_extra_regs;
#else	/* CONFIG_OSFMACH3 */
	int s = _disable_interrupts();
#endif	/* CONFIG_OSFMACH3 */

	while ((signr = current->signal & mask)) {
#if 0
		signr = ffz(~signr);  /* Compute bit # */
#else
		for (bitno = 0;  bitno < 32;  bitno++)
		{
			if (signr & (1<<bitno)) break;
		}
		signr = bitno;
#endif
		current->signal &= ~(1<<signr);  /* Clear bit */
		sa = current->sig->action + signr;
		signr++;
		if ((current->flags & PF_PTRACED) && signr != SIGKILL) {
			current->exit_code = signr;
			current->state = TASK_STOPPED;
			notify_parent(current);
			schedule();
			if (!(signr = current->exit_code))
				continue;
			current->exit_code = 0;
			if (signr == SIGSTOP)
				continue;
			if (_S(signr) & current->blocked) {
				current->signal |= _S(signr);
				continue;
			}
			sa = current->sig->action + signr - 1;
		}
		if (sa->sa_handler == SIG_IGN) {
			if (signr != SIGCHLD)
				continue;
			/* check for SIGCHLD: it's special */
			while (sys_waitpid(-1,NULL,WNOHANG) > 0)
				/* nothing */;
			continue;
		}
		if (sa->sa_handler == SIG_DFL) {
			if (current->pid == 1)
				continue;
			switch (signr) {
			case SIGCONT: case SIGCHLD: case SIGWINCH:
				continue;

			case SIGSTOP: case SIGTSTP: case SIGTTIN: case SIGTTOU:
				if (current->flags & PF_PTRACED)
					continue;
				current->state = TASK_STOPPED;
				current->exit_code = signr;
				if (!(current->p_pptr->sig->action[SIGCHLD-1].sa_flags &
						SA_NOCLDSTOP))
					notify_parent(current);
				schedule();
				continue;

			case SIGQUIT: case SIGILL: case SIGTRAP:
			case SIGIOT: case SIGFPE: case SIGSEGV:
				if (current->binfmt && current->binfmt->core_dump) {
					if (current->binfmt->core_dump(signr, regs))
						signr |= 0x80;
				}
				/* fall through */
			default:
				current->signal |= _S(signr & 0x7f);
				do_exit(signr);
			}
		}
		/*
		 * OK, we're invoking a handler
		 */
#ifdef	CONFIG_OSFMACH3
		if (extra_regs->trap == 0x0c00)
#else	/* CONFIG_OSFMACH3 */
		if ((int)regs->orig_gpr3 >= 0)
#endif	/* CONFIG_OSFMACH3 */
		{
#ifdef	CONFIG_OSFMACH3
			if ((int)extra_regs->result == -ERESTARTNOHAND ||
			   ((int)extra_regs->result == -ERESTARTSYS &&
			    !(sa->sa_flags & SA_RESTART)))
				(int)extra_regs->result = -EINTR;
#else	/* CONFIG_OSFMACH3 */
			if ((int)regs->result == -ERESTARTNOHAND ||
			   ((int)regs->result == -ERESTARTSYS && !(sa->sa_flags & SA_RESTART)))
				(int)regs->result = -EINTR;
#endif	/* CONFIG_OSFMACH3 */
		}
		handler_signal |= 1 << (signr-1);
		mask &= ~sa->sa_mask;
	}

#ifdef	CONFIG_OSFMACH3
	if (extra_regs->trap == 0x0c00 &&
	    ((int)extra_regs->result == -ERESTARTNOHAND ||
	     (int)extra_regs->result == -ERESTARTSYS ||
	     (int)extra_regs->result == -ERESTARTNOINTR)) {
		regs->gpr[3] = extra_regs->orig_gpr3;
		regs->nip -= 4; /* Back up & retry system call */
		extra_regs->result = 0;
	}
#else	/* CONFIG_OSFMACH3 */
	if ((int)regs->orig_gpr3 >= 0 &&
	    ((int)regs->result == -ERESTARTNOHAND ||
	     (int)regs->result == -ERESTARTSYS ||
	     (int)regs->result == -ERESTARTNOINTR)) {
		regs->gpr[3] = regs->orig_gpr3;
		regs->nip -= 4; /* Back up & retry system call */
		regs->result = 0;
	}
#endif	/* CONFIG_OSFMACH3 */

	if (!handler_signal)		/* no handler will be called - return 0 */
	{
#ifndef	CONFIG_OSFMACH3
		_enable_interrupts(s);
#endif	/* CONFIG_OSFMACH3 */
		return 0;
	}
	nip = regs->nip;
	frame = (unsigned long *) regs->gpr[1];

#ifdef	CONFIG_OSFMACH3
	/*
	 * First compute the size of the stuff to push on the stack
	 * and check that the user memory is there.
	 */
	frame_size = 2 + (sizeof (*regs) / sizeof (long));
	for (mask = 1; mask; mask += mask) {
		if (mask > handler_signal)
			break;
		if (!(mask & handler_signal))
			continue;
		frame_size +=
			sizeof (struct sigcontext_struct) / sizeof (long);
	}
	if (verify_area(VERIFY_WRITE,
			frame - frame_size,
			frame_size * sizeof (long))) {
		do_exit(SIGSEGV);
	}
#endif	/* CONFIG_OSFMACH3 */

	/* Build trampoline code on stack */
	frame -= 2;
	trampoline = frame;
	put_fs_long(0x38007777, trampoline);      /* li r0,0x7777 */
	put_fs_long(0x44000002, trampoline+1);    /* sc           */
	frame -= sizeof(*regs) / sizeof(long);
	regs_ptr = frame;

#ifdef	CONFIG_OSFMACH3
	flush_instruction_cache_range(current->mm,
				      (unsigned long) trampoline,
				      (unsigned long) (trampoline + 2));
	/*
	 * Copy the regular registers (common to Mach and Linux).
	 */
	memcpy_tofs(regs_ptr, regs,
		    ((int)(&regs->pt_extra_regs)) -
		    ((int)(regs)));
	memcpy_tofs(&((struct pt_regs *)regs_ptr)->pt_extra_regs + 1,
		    &extra_regs->pt_extra_regs + 1,
		    (sizeof (*regs) -
		     (((int)(&regs->pt_extra_regs+1)) - ((int)(regs)))));
#else	/* CONFIG_OSFMACH3 */
	memcpy_tofs(regs_ptr, regs, sizeof(*regs));
#endif	/* CONFIG_OSFMACH3 */
	signr = 1;
	sa = current->sig->action;
#ifdef	CONFIG_OSFMACH3
	sc = NULL;
#endif	/* CONFIG_OSFMACH3 */		
	for (mask = 1 ; mask ; sa++,signr++,mask += mask) {
		if (mask > handler_signal)
			break;
		if (!(mask & handler_signal))
			continue;
		frame -= sizeof(struct sigcontext_struct) / sizeof(long);
#ifdef DEBUG_SIGNALS
#ifdef	CONFIG_OSFMACH3
printk("Send signal: %ld, frame: %p\n", signr, frame);
#else	/* CONFIG_OSFMACH3 */
printk("Send signal: %d, frame: %x\n", signr, frame);
#endif	/* CONFIG_OSFMACH3 */
#endif
		sc = (struct sigcontext_struct *)frame;
		nip = (unsigned long) sa->sa_handler;
		if (sa->sa_flags & SA_ONESHOT)
			sa->sa_handler = NULL;
		put_fs_long(nip, &sc->handler);
		put_fs_long(oldmask, &sc->oldmask); /* was current->blocked */
#ifdef	CONFIG_OSFMACH3
		put_fs_long((long)regs_ptr, &sc->regs);
#else	/* CONFIG_OSFMACH3 */
		put_fs_long(regs_ptr, &sc->regs);
#endif	/* CONFIG_OSFMACH3 */
		put_fs_long(signr, &sc->signal);
		current->blocked |= sa->sa_mask;
		regs->gpr[3] = signr;
		regs->gpr[4] = (unsigned long)regs_ptr;
	}
	regs->link = (unsigned long)trampoline;
	regs->nip = nip;
	regs->gpr[1] = (unsigned long)sc - STACK_FRAME_OVERHEAD;
#ifdef DEBUG_SIGNALS
#ifdef	CONFIG_OSFMACH3
printk("Signal - PC: %lx, SP: %lx, SC: %p\n", regs->nip, regs->gpr[1], sc);
#else	/* CONFIG_OSFMACH3 */
printk("Signal - PC: %x, SP: %x, SC: %x\n", regs->nip, regs->gpr[1], sc);
#endif	/* CONFIG_OSFMACH3 */
#endif
	/* The DATA cache must be flushed here to insure coherency */
	/* between the DATA & INSTRUCTION caches.  Since we just */
	/* created an instruction stream using the DATA [cache] space */
	/* and since the instruction cache will not look in the DATA */
	/* cache for new data, we have to force the data to go on to */
	/* memory and flush the instruction cache to force it to look */
	/* there.  The following function performs this magic */
	flush_instruction_cache();
#ifndef	CONFIG_OSFMACH3
	_enable_interrupts(s);
#endif	/* CONFIG_OSFMACH3 */
	return 1;
}
