/*
 * @OSF_FREE_FREE_COPYRIGHT@
 * 
 */
/*
 * HISTORY
 * $Log: osfmach3.c,v $
 * Revision 1.1.2.3  1996/09/19  14:41:31  barbou
 * 	Changed OSFMACH3 conditionals to CONFIG_OSFMACH3 and
 * 	make sure that the configuration header file is included.
 * 	[1996/09/19  14:33:39  barbou]
 *
 * Revision 1.1.2.2  1996/09/10  18:02:19  barbou
 * 	Fixed osfmach3_proc_dir_entry: specify the inode ops.
 * 	[1996/09/10  18:01:15  barbou]
 * 
 * Revision 1.1.2.1  1996/09/09  16:44:17  barbou
 * 	Fixed proc_register call.
 * 	[96/09/06            barbou]
 * 
 * 	Fixed typo.
 * 	[96/08/22            barbou]
 * 
 * 	Added proc_osfmach3_init() to register the osfmach3 files.
 * 	[96/08/22            barbou]
 * 
 * 	Adapted to the new procfs generic structure.
 * 	[96/08/22            barbou]
 * 
 * 	Created.
 * 	[1996/08/22  11:03:50  barbou]
 * 
 * $EndLog$
 */

#include <linux/autoconf.h>

#ifdef	CONFIG_OSFMACH3

#include <mach/mach_interface.h>
#include <mach/mach_host.h>

#include <osfmach3/mach3_debug.h>

#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/stat.h>
#include <linux/fcntl.h>
#include <linux/config.h>
#include <linux/mm.h>

#include <asm/segment.h>

/* external declarations */
extern mach_port_t      host_port;

/* forward references */
static int proc_readosfmach3(struct inode *inode,
			     struct file *file,
			     char *buf,
			     int count);

static struct file_operations proc_osfmach3_operations = {
	NULL,			/* lseek - default */
	proc_readosfmach3,	/* read - bad */
	NULL,			/* write - bad */
	proc_readdir,		/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	NULL			/* can't fsync */
};

/*
 * proc directories can do almost nothing..
 */
struct inode_operations proc_osfmach3_inode_operations = {
	&proc_osfmach3_operations,	/* osfmach3 directory file-ops */
	NULL,			/* create */
	proc_lookup,		/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL			/* permission */
};

static int proc_readosfmach3(
	struct inode *inode,
	struct file *file,
	char *buf,
	int count)
{
	char * page;
	int length;
	unsigned int ino;
	kern_return_t	kr;

	if (count < 0)
		return -EINVAL;
	if (!(page = (char*) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;
	ino = inode->i_ino;

	switch (ino) 
	{
	case PROC_OSFMACH3_VERSION:

		kr = host_kernel_version(host_port, page);
		
		if (kr != KERN_SUCCESS) {
			MACH3_DEBUG(2, kr,
				    ("proc_readosfmach3: host_kernel_version"));
		}
			
		length = strlen(page);
		break;
	default:
		free_page((unsigned long) page);
		return -EBADF;
	}
		
	if (file->f_pos >= length) {
		free_page((unsigned long) page);
		return 0;
	}
	if (count + file->f_pos > length)
		count = length - file->f_pos;

	/*
	 *	Copy the bytes
	 */
	memcpy_tofs(buf, page + file->f_pos, count);

	free_page((unsigned long) page);

	file->f_pos += count;	/* Move down the file */

	return count;	/* all transfered correctly */
}

static struct proc_dir_entry osfmach3_proc_dir_entry = {
	PROC_OSFMACH3_VERSION, 7, "version",
	S_IFREG | S_IRUGO, 1, 0, 0,
	0, &proc_osfmach3_inode_operations,
};

void
proc_osfmach3_init(void)
{
	proc_register(&proc_osfmach3, &osfmach3_proc_dir_entry);
}

#endif	/* CONFIG_OSFMACH3 */
