/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 * 
 */
/*
 * MkLinux
 */

#include <mach_kgdb.h>
#include <mach_kdb.h>
#include <debug.h>
#include <mach_debug.h>
#include <kern/misc_protos.h>
#include <kern/assert.h>

#include <kgdb/gdb_defs.h>
#include <kgdb/kgdb_defs.h>	/* For kgdb_printf */

#include <ppc/spl.h>
#include <ppc/proc_reg.h>
#include <ppc/misc_protos.h>
#include <ppc/trap.h>
#include <ppc/exception.h>
#include <ppc/POWERMAC/powermac.h>
#include <ppc/POWERMAC/interrupts.h>
#include <ppc/POWERMAC/powermac_pci.h>

/* PDM interrupt support is very limited, a certain amount of
 * masking can be done using the ICR but we don't do this yet.
 * SPLs are currently all or nothing.
 */


/* Two identical functions with different names to allow single stepping
 * of pdm_spl whilst in kdb. Define contents in a macro so that it stays
 * identical!
 */

#define SPL_FUNCTION(NAME)						\
static spl_t NAME(spl_t lvl);						\
spl_t NAME(spl_t lvl)							\
{									\
	spl_t old_level;						\
									\
	old_level = get_interrupt_level();				\
									\
	if (lvl == SPLLO) {						\
		cpu_data[cpu_number()].interrupt_level = lvl;		\
		interrupt_enable();					\
	} else {							\
		interrupt_disable();					\
		cpu_data[cpu_number()].interrupt_level = lvl;		\
	}								\
									\
	/* Make sure that the write completes before we continue */	\
									\
	sync();								\
									\
	return old_level;						\
}

SPL_FUNCTION(pci_spl)

#if	MACH_KDB
SPL_FUNCTION(pci_db_spl)
#endif	/* MACH_KDB */

static void	pci_via1_interrupt (int device, void *ssp);

#define NVIA1_INTERRUPTS 7
struct powermac_interrupt pci_via1_interrupts[NVIA1_INTERRUPTS] = {
	{ 0,	0,	-1},			/* Cascade */
	{ 0,	0,	PMAC_DEV_HZTICK},
	{ 0,	0,	PMAC_DEV_CUDA},		
	{ 0,	0, 	-1},			/* VIA Data */
	{ 0,	0, 	PMAC_DEV_PMU},		/* VIA CLK Source */
	{ 0,	0,	PMAC_DEV_TIMER2},
	{ 0,	0,	PMAC_DEV_TIMER1}
};

#define NPCI_INTERRUPTS 32

/* This structure is little-endian formatted... */

struct powermac_interrupt  powermac_pci_interrupts[NPCI_INTERRUPTS] = {
	{ 0, 	0, PMAC_DEV_CARD4},	  /* Bit 24 - External Int 4 */
	{ 0, 	0, PMAC_DEV_CARD5},	  /* Bit 25 - External Int 5 */
	{ 0, 	0, PMAC_DEV_CARD6},	  /* Bit 26 - External Int 6 */
	{ 0, 	0, PMAC_DEV_CARD7},	  /* Bit 27 - External Int 7 */
	{ 0, 	0, PMAC_DEV_CARD8},	  /* Bit 28 - External Int 8 */
	{ 0, 	0, PMAC_DEV_CARD9},	  /* Bit 29 - External Int 9 */
	{ 0, 	0, PMAC_DEV_CARD10},	  /* Bit 30 - External Int 10 */
	{ 0, 	0, -1},			  /* Bit 31 - Reserved */
	{ 0,	0, PMAC_DEV_SCC_B},	  /* Bit 16 - SCC Channel B */
	{ 0,	0, PMAC_DEV_AUDIO}, 	  /* Bit 17 - Audio */
	{ pci_via1_interrupt, 	0, -1},	  /* Bit 18 - VIA (cuda) */
	{ 0,	0, PMAC_DEV_FLOPPY},	  /* Bit 19 - SwimIII/Floppy */
	{ 0, 	0, PMAC_DEV_CARD0},	  /* Bit 20 - External Int 0 */
	{ 0, 	0, PMAC_DEV_CARD1},	  /* Bit 21 - External Int 1 */
	{ 0, 	0, PMAC_DEV_CARD2},	  /* Bit 22 - External Int 2 */
	{ 0, 	0, PMAC_DEV_CARD3},	  /* Bit 23 - External Int 3 */
	{ 0,	0, PMAC_DMA_AUDIO_OUT},	  /* Bit 8 - DMA Audio Out */
	{ 0,	0, PMAC_DMA_AUDIO_IN},	  /* Bit 9 - DMA Audio In */
	{ 0,	0, PMAC_DMA_SCSI1},	  /* Bit 10 - DMA SCSI 1 */
	{ 0,	0, -1},			  /* Bit 11 - Reserved */
	{ 0,	0, PMAC_DEV_SCSI0},	  /* Bit 12 - SCSI 0 */
	{ 0,	0, PMAC_DEV_SCSI1},	  /* Bit 13 - SCSI 1 */
	{ 0,	0, PMAC_DEV_ETHERNET},	  /* Bit 14 - MACE/Ethernet */
	{ 0,	0, PMAC_DEV_SCC_A},	  /* Bit 15 - SCC Channel A */
	{ 0,	0, PMAC_DMA_SCSI0},	  /* Bit 0 - DMA SCSI 0 */
	{ 0,	0, PMAC_DMA_FLOPPY},	  /* Bit 1 - DMA Floppy */
	{ 0,	0, PMAC_DMA_ETHERNET_TX}, /* Bit 2 - DMA Ethernet Tx */
	{ 0,	0, PMAC_DMA_ETHERNET_RX}, /* Bit 3 - DMA Ethernet Rx */
	{ 0,	0, PMAC_DMA_SCC_A_TX},	  /* Bit 4 - DMA SCC Channel A TX */
	{ 0,	0, PMAC_DMA_SCC_A_RX},	  /* Bit 5 - DMA SCC Channel A RX */
	{ 0,	0, PMAC_DMA_SCC_B_TX},	  /* Bit 6 - DMA SCC Channel B TX */
	{ 0,	0, PMAC_DMA_SCC_B_RX}	  /* Bit 7 - DMA SCC Channel B RX */
};

static void	pci_interrupt(int type,
			struct ppc_saved_state *ssp,
			unsigned int dsisr, unsigned int dar);

/* Reset the hardware interrupt control */
void
pci_interrupt_initialize(void)
{
	pmac_register_int	= pci_register_int;
	platform_interrupt	= pci_interrupt;
	platform_spl		= pci_spl;
#if	MACH_KDB
	platform_db_spl		= pci_db_spl;
#endif	/* MACH_KDB */

	*PCI_INTERRUPT_MASK  = 0;	   /* Disable all interrupts */
	eieio();
	*PCI_INTERRUPT_CLEAR = 0xffffffff; /* Clear pending interrupts */
	eieio();
	*PCI_INTERRUPT_MASK  = 0;	   /* Disable all interrupts */
	eieio();
	*(v_u_char *)PCI_VIA1_PCR               = 0x00; eieio();
        *(v_u_char *)PCI_VIA1_IER               = 0x00; eieio();
        *(v_u_char *)PCI_VIA1_IFR               = 0x7f; eieio();
}

static int
pci_find_entry(int device,
	       struct powermac_interrupt **handler,
	       int nentries)
{
	int	i;

	for (i = 0; i < nentries; i++, (*handler)++)
		if ( (*handler)->i_device == device)
			return i;

	*handler = NULL;
	return 0;
}

void
pci_register_int(int device, spl_t level, void (*handler)(int, void *))
{
	int	i;
	struct powermac_interrupt	*p;

	/* Check primary interrupts */
	p = powermac_pci_interrupts;
	i = pci_find_entry(device, &p, NPCI_INTERRUPTS);
	if (p) {
		if (p->i_handler) {
			panic("pci_register_int: "
			      "Interrupt %d already taken!? ", device);
		} else {
			p->i_handler = handler;
			p->i_level = level;
			*PCI_INTERRUPT_MASK |= (1<<i);
			eieio();
			return;
		}
		return;
	}
	/* Check cascaded interrupts */
	p = pci_via1_interrupts;
	i = pci_find_entry(device, &p, NVIA1_INTERRUPTS);
	if (p) {
		if (p->i_handler) {
			panic("pci_register_int: "
				"Interrupt %d already taken!? ", device);
		} else {
			p->i_handler = handler;
			p->i_level = level;

			*((v_u_char *) PCI_VIA1_IER) |= (1 << i);
			eieio();
			*PCI_INTERRUPT_MASK |= (1<<10);	/* enable */
			eieio();
			return;
		}
		return;
	}

	panic("pci_register_int: Interrupt %d not found", device);
}

static void
pci_interrupt( int type, struct ppc_saved_state *ssp,
	       unsigned int dsisr, unsigned int dar)
{
	unsigned long int		bit, irq;
	struct powermac_interrupt	*handler;

	irq = *PCI_INTERRUPT_EVENTS; eieio();
	*PCI_INTERRUPT_CLEAR = irq;	/* Clear out interrupts */
	eieio();

	handler = powermac_pci_interrupts;

	for (bit = 0; bit < NPCI_INTERRUPTS ; bit++, handler++) {
		if (irq & (1<<bit))
			if (handler->i_handler)
				handler->i_handler(handler->i_device, ssp);
			else
				printf("{PCI INT %d}", bit);
	}
}

void
pci_via1_interrupt(int device, void *ssp)
{
	register unsigned char intbits, bit;
	struct powermac_interrupt	*handler;

	intbits = via_reg(PCI_VIA1_IFR); eieio();/* get interrupts pending */
	intbits &= via_reg(PCI_VIA1_IER); eieio();	/* only care about enabled */

	if (intbits == 0)
		return;

	/*
	 * Unflag interrupts we're about to process.
	 */
	via_reg(PCI_VIA1_IFR) = intbits;
	eieio();

	handler = pci_via1_interrupts;
	for (bit = 0; bit < NVIA1_INTERRUPTS ; bit++, handler++) {
		if(intbits & (1<<bit)) {
			if (handler->i_handler)  
				handler->i_handler(handler->i_device, ssp);
		}
	}
}

