/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux
 */

#include <cpus.h>

#include <ppc/asm.h>
#include <ppc/proc_reg.h>
#include <cpus.h>
#include <assym.s>
#include <mach_debug.h>
#include <mach/ppc/vm_param.h>

/*
 * extern void flush_cache(vm_offset_t pa, unsigned count);
 *
 * flush_cache takes a physical address and count to flush, thus
 * must not be called for multiple virtual pages.
 *
 * it flushes the data cache and invalidates the instruction
 * cache for the address range in question
 */

ENTRY(flush_cache, TAG_NO_FRAME_USED)

	/* Switch off data translations */
	mfmsr	r6
	rlwinm	r7,	r6,	0,	MSR_DR_BIT+1,	MSR_DR_BIT-1
	mtmsr	r7
	isync

	/* Check to see if the address is aligned. */
	add	r8, ARG0,ARG1
	andi.	r8,r8,(CACHE_LINE_SIZE-1)
	beq-	.L_flush_check
	addi	ARG1,ARG1,CACHE_LINE_SIZE

.L_flush_check:
	/* Make ctr hold count of how many times we should loop */
	srwi.	r8,	ARG1,	CACHE_LINE_POW2
	/* If less than 32 to do, jump over loop */
	beq- .L_flush_cache_done
	mtctr	r8

.L_flush_loop:
	subic	ARG1,	ARG1,	CACHE_LINE_SIZE
	dcbf	ARG0,	ARG1
	icbi	ARG0,	ARG1
	bdnz	.L_flush_loop

.L_flush_cache_done:
	dcbf	0,ARG0
	icbi	0,ARG0

	sync			/* Finish physical writes */
	mtmsr	r6		/* Restore original translations */
	isync			/* Ensure data translations are on */

	blr

/*
 * extern void flush_cache_v(vm_offset_t pa, unsigned count);
 *
 * flush_cache_v takes a virtual address and count to flush, thus
 * can be called for multiple virtual pages.
 *
 * it flushes the data cache and invalidates the instruction
 * cache for the address range in question
 */

ENTRY(flush_cache_v, TAG_NO_FRAME_USED)

	/* Check to see if the address is aligned. */
	add	r8, ARG0,ARG1
	andi.	r8,r8,(CACHE_LINE_SIZE-1)
	beq-	.L_flushv_check
	addi	ARG1,ARG1,CACHE_LINE_SIZE

.L_flushv_check:
	/* Make ctr hold count of how many times we should loop */
	srwi.	r8,	ARG1,	CACHE_LINE_POW2
	/* If less than 32 to do, jump over loop */
	beq- .L_flush_cache_v_done
	mtctr	r8

.L_flushv_loop:
	subic	ARG1,	ARG1,	CACHE_LINE_SIZE
	dcbf	ARG0,	ARG1
	icbi	ARG0,	ARG1
	bdnz	.L_flushv_loop

.L_flush_cache_v_done:
	xor	ARG1,ARG1,ARG1
	dcbf	0,ARG0
	icbi	0,ARG0

	sync		/* make sure flushes have completed */
	isync

	blr

/*
 * extern void invalidate_cache_v(vm_offset_t pa, unsigned count);
 */

ENTRY(invalidate_cache_v, TAG_NO_FRAME_USED)
	/* Check to see if the address is aligned. */
	add	r8, ARG0,ARG1
	andi.	r8,r8,(CACHE_LINE_SIZE-1)
	beq-	.L_invalidate_checkv
	addi	ARG1,ARG1,CACHE_LINE_SIZE

.L_invalidate_checkv:
	/* Make ctr hold count of how many times we should loop */
	srwi.	r8,	ARG1,	CACHE_LINE_POW2
	/* If less than 32 to do, jump over loop */
	beq- .L_invalidate_cache_v_done
	mtctr	r8

.L_invalidatev_loop:
	subic	ARG1,	ARG1,	CACHE_LINE_SIZE
	dcbi	ARG0,	ARG1
	icbi	ARG0,	ARG1
	bdnz	.L_invalidatev_loop

.L_invalidate_cache_v_done:
	dcbi	0,ARG0
	icbi	0,ARG0

	sync		/* make sure invalidates have completed */
	isync

	blr


/*
 * void sync_tlb(void)
 *	Synchronize TLBs on MP machines.
 */
#if	NCPUS > 1
ENTRY(sync_tlb, TAG_NO_FRAME_USED)
	tlbsync
	sync
	blr
#endif	/* NCPUS > 1 */
