/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux
 */

/* Miscellaneous constants and structures used by the exception
 * handlers
 */

#ifndef _PPC_EXCEPTION_H_
#define _PPC_EXCEPTION_H_

#ifndef ASSEMBLER

#include <cpus.h>
#include <mach_kgdb.h>

#include <mach/machine/vm_types.h>

/* When an exception is taken, this info is accessed via sprg0 */
struct per_proc_info {
	unsigned int	cpu_number;
	vm_offset_t  	istackptr;
	vm_offset_t  	intstack_top_ss;
#if	MACH_KGDB
	vm_offset_t  	gdbstackptr;
	vm_offset_t  	gdbstack_top_ss;
#endif	/* MACH_KGDB */

	unsigned int save_cr;
	unsigned int save_srr0;
	unsigned int save_srr1;
	unsigned int save_exception_type;
	unsigned int phys_exception_handlers;
	unsigned int virt_per_proc_info; /* virt addr for our CPU */
	unsigned int active_kloaded;	/* pointer to active_kloaded[CPU_NO] */
	unsigned int cpu_data;		/* pointer to cpu_data[CPU_NO] */
	unsigned int active_stacks;	/* pointer to active_stacks[CPU_NO] */
	unsigned int need_ast;		/* pointer to need_ast[CPU_NO] */
	unsigned int fpu_pcb;		/* pcb owning the fpu on this cpu */
};

extern struct per_proc_info per_proc_info[NCPUS];

extern char *trap_type[];

static __inline__ struct per_proc_info * current_proc_info(void)
{
	register struct per_proc_info *ppi;
							   
	__asm__ volatile ("mr %0, 2" : "=r" (ppi));
	return ppi;
}

#endif /* ndef ASSEMBLER */

#define T_VECTOR_SIZE		4		/* function pointer size */

/* Hardware exceptions */

#define T_INVALID		(0x00 * T_VECTOR_SIZE)
#define T_RESET			(0x01 * T_VECTOR_SIZE)
#define T_MACHINE_CHECK		(0x02 * T_VECTOR_SIZE)
#define T_DATA_ACCESS		(0x03 * T_VECTOR_SIZE)
#define T_INSTRUCTION_ACCESS	(0x04 * T_VECTOR_SIZE)
#define T_INTERRUPT		(0x05 * T_VECTOR_SIZE)
#define T_ALIGNMENT		(0x06 * T_VECTOR_SIZE)
#define T_PROGRAM		(0x07 * T_VECTOR_SIZE)
#define T_FP_UNAVAILABLE	(0x08 * T_VECTOR_SIZE)
#define T_DECREMENTER		(0x09 * T_VECTOR_SIZE)
#define T_IO_ERROR		(0x0a * T_VECTOR_SIZE)
#define T_RESERVED		(0x0b * T_VECTOR_SIZE)
#define T_SYSTEM_CALL		(0x0c * T_VECTOR_SIZE)
#define T_TRACE			(0x0d * T_VECTOR_SIZE)
#define T_FP_ASSIST		(0x0e * T_VECTOR_SIZE)
#define T_INSTRUCTION_BKPT	(0x13 * T_VECTOR_SIZE)
#define T_SYSTEM_MANAGEMENT	(0x14 * T_VECTOR_SIZE)

#define T_RUNMODE_TRACE		(0x20 * T_VECTOR_SIZE) /* 601 only */

/* software exceptions */

#define T_AST			(0x100 * T_VECTOR_SIZE) 
#define T_MAX			T_RUNMODE_TRACE /* Maximum exception no */

#endif /* _PPC_EXCEPTION_H_ */
