/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 * 
 */
/*
 * MkLinux
 */

#include <cpus.h>
#include <mach_assert.h>
#include <mach_ldebug.h>
#include <mach_rt.h>

#include <kern/etap_options.h>
	
#include <ppc/asm.h>
#include <assym.s>

#if	NCPUS > 1
#include <ppc/POWERMAC/mp/mp.h>
#endif	/* NCPUS > 1 */

#define PROLOG(space)							\
	stwu	r1,	-((space)+16)(r1);				\
	mflr	r0		;					\
	stw	r0,	((space)+20)(r1);				\
	stw	r3,	12(r1)
#define EPILOG								\
	lwz	r11,	0(r1)	;					\
	lwz	r0,	4(r11)	;					\
	mtlr	r0		;					\
	mr	r1,	r11
	
#if	MACH_KDB || MACH_KGDB
	
#define OPTIONAL_PROLOG		PROLOG(4)
#define OPTIONAL_EPILOG		EPILOG
	
#else	/* MACH_KDB || MACH_KGDB */
	
#define OPTIONAL_PROLOG
#define OPTIONAL_EPILOG
	
#endif	/* MACH_KDB || MACH_KGDB */

#if	MACH_RT
#error DISABLE_PREEMPTION not implemented
#error ENABLE_PREEMPTION not implemented
#else	/* MACH_RT */
#define DISABLE_PREEMPTION()
#define ENABLE_PREEMPTION()
#endif	/* MACH_RT */
	
#if	MACH_LDEBUG
/*
 * Routines for general lock debugging.
 */

/*
 * Checks for expected lock types and calls "panic" on
 * mismatch.  Detects calls to Mutex functions with
 * type simplelock and vice versa.
 */
#define	CHECK_MUTEX_TYPE()						\
	lwz	r10,	MUTEX_TYPE(r3);					\
	cmpwi	r10,	MUTEX_TAG;					\
	beq+	1f;							\
	lis	r3,	not_a_mutex@ha;					\
	ori	r3,	r3,	not_a_mutex@l;				\
	bl	EXT(panic)	;					\
	lwz	r3,	12(r1)	;					\
1:
	
	.data
not_a_mutex:
	.string	"not a mutex!\n"
	.text

#define CHECK_SIMPLE_LOCK_TYPE()					\
	lwz	r10,	SLOCK_TYPE(r3);					\
	cmpwi	r10,	USLOCK_TAG;					\
	beq+	1f;							\
	lis	r3,	not_a_slock@ha;					\
	ori	r3,	r3,	not_a_slock@l;				\
	bl	EXT(panic)	;					\
	lwz	r3,	12(r1)	;					\
1:
	
	.data
not_a_slock:
	.string	"not a simple lock!\n"
	.text

/*
 * If one or more simplelocks are currently held by a thread,
 * an attempt to acquire a mutex will cause this check to fail
 * (since a mutex lock may context switch, holding a simplelock
 * is not a good thing).
 */
#if	MACH_RT
#error CHECK_PREEMPTION_LEVEL not implemented
#else	/* MACH_RT */
#define CHECK_PREEMPTION_LEVEL()
#endif	/* MACH_RT */
	
#define CHECK_NO_SIMPLELOCKS()						\
	lwz	r10,	PP_CPU_DATA(r2);				\
	lwz	r10,	CPU_SIMPLE_LOCK_COUNT(r10);			\
	cmpwi	r10,	0;						\
	beq+	1f;							\
	lis	r3,	simple_locks_held@ha;				\
	ori	r3,	r3,	simple_locks_held@l;			\
	bl	EXT(panic)	;					\
	lwz	r3,	12(r1)	;					\
1:									\
	CHECK_PREEMPTION_LEVEL()
	
	.data
simple_locks_held:
	.string	"simple locks held!\n";
	.text

/* 
 * Verifies return to the correct thread in "unlock" situations.
 */
#define CHECK_THREAD(thread_offset)					\
	lwz	r10,	PP_CPU_DATA(r2);				\
	lwz	r10,	CPU_ACTIVE_THREAD(r10);				\
	cmpwi	r10,	0	;					\
	beq-	1f;							\
	lwz	r9,	thread_offset(r3);				\
	cmpw	r9,	r10	;					\
	beq+	1f;							\
	lis	r3,	wrong_thread@ha;				\
	ori	r3,	r3,	wrong_thread@l;				\
	bl	EXT(panic)	;					\
	lwz	r3,	12(r1)	;					\
1:
	
	.data
wrong_thread:
	.string	"wrong thread!\n"
	.text

#define CHECK_MYLOCK(thread_offset)					\
	lwz	r10,	PP_CPU_DATA(r2);				\
	lwz	r10,	CPU_ACTIVE_THREAD(r10);				\
	cmpwi	r10,	0	;					\
	beq-	1f;							\
	lwz	r9,	thread_offset(r3);				\
	cmpw	r9,	r10	;					\
	bne+	1f;							\
	lis	r3,	mylock_attempt@ha;				\
	ori	r3,	r3,	mylock_attempt@l;			\
	bl	EXT(panic)	;					\
	lwz	r3,	12(r1)	;					\
1:	
	
	.data
mylock_attempt:
	.string	"mylock attempt!\n"
	.text

#else	/* MACH_LDEBUG */

#define CHECK_MUTEX_TYPE()
#define CHECK_SIMPLE_LOCK_TYPE()
#define CHECK_THREAD(thread_offset)
#define CHECK_NO_SIMPLELOCKS()
#define CHECK_MYLOCK(thread_offset)

#endif	/* MACH_LDEBUG */
	
/*
 *      void hw_lock_init(hw_lock_t)
 *
 *      Initialize a hardware lock.
 * TODO NMGS Manual talks about possible races if write to same cache line?
 * TODO NMGS should these locks take up a whole cache line? See HP code
 */

ENTRY(hw_lock_init, TAG_NO_FRAME_USED)

	OPTIONAL_PROLOG
	li	r0,	0		/* set lock to free == 0 */
	stw	r0,	0(ARG0)
	OPTIONAL_EPILOG
	blr
	
/*
 *      void hw_lock_lock(hw_lock_t)
 *
 *      Acquire lock, spinning until it becomes available.
 *      MACH_RT:  also return with preemption disabled. TODO NMGS
 */

ENTRY(hw_lock_lock, TAG_NO_FRAME_USED)

	PROLOG(0)
	li	ARG1,	1		/* value to be stored... 1==taken*/
.L_lock_lock_retry:
	DISABLE_PREEMPTION()
	lwarx	ARG2,	0,ARG0		/* Ld from addr of ARG0 and reserve */
	cmpwi	ARG2,	0		/* TEST... */
	bne-	.L_lock_lock_loop	/* loop if BUSY (predict not busy) */
	
	stwcx.  ARG1,	0,ARG0		/* And SET (if reserved) */
	bne-	.L_lock_lock_loop	/* If set failed, loop back */
	isync
	EPILOG
	blr
.L_lock_lock_loop:
	ENABLE_PREEMPTION()
	lwz	ARG2,	0(ARG0)
	cmpwi	ARG2,	0
	bne+	.L_lock_lock_loop
	b	.L_lock_lock_retry

/*
 *      void hw_lock_unlock(hw_lock_t)
 *
 *      Unconditionally release lock.
 *      MACH_RT:  release preemption level. TODO NMGS
 */

	.text

ENTRY(hw_lock_unlock, TAG_NO_FRAME_USED)

	PROLOG(0)
	sync				/* Flush writes done under lock */
	li	r0,	0		/* set lock to free == 0 */
	stw	r0,	0(ARG0)
	ENABLE_PREEMPTION()
	EPILOG
	blr

/*
 *      unsigned int hw_lock_try(hw_lock_t)
 *
 *      try to acquire spin-lock. Return success (1) or failure (0)
 *      MACH_RT:  returns with preemption disabled on success. TODO NMGS
 *
 */
ENTRY(hw_lock_try,TAG_NO_FRAME_USED)

	PROLOG(0)
	li	ARG1,	1		/* value to be stored... 1==taken */
	DISABLE_PREEMPTION()
.L_lock_try_loop:	
	lwarx	ARG2,	0,ARG0		/* Ld from addr of arg and reserve */

	cmpwi	ARG2,	0		/* TEST... */
	bne-	.L_lock_try_failed	/* branch if taken. Predict free */
	
	stwcx.  ARG1,	0,ARG0
		/* And SET (if reserved) */
	bne-	.L_lock_try_loop	/* If set failed, loop back */
	isync

	li	ARG0,	1		/* SUCCESS - lock was free */
	EPILOG
	blr

.L_lock_try_failed:
	ENABLE_PREEMPTION()
	li	ARG0,	0		/* FAILURE - lock was taken */
	EPILOG
	blr

/*
 *      unsigned int hw_lock_held(hw_lock_t)
 *
 *      Return 1 if lock is held
 *      MACH_RT:  doesn't change preemption state.
 *      N.B.  Racy, of course.
 *
 */
ENTRY(hw_lock_held,TAG_NO_FRAME_USED)

	OPTIONAL_PROLOG
	lwz	ARG0, 0(ARG0)		/* Return value of lock */
	OPTIONAL_EPILOG
	blr

/*
 *	void mutex_init(mutex_t* l, etap_event_t etap)
 */

ENTRY(mutex_init,TAG_NO_FRAME_USED)

	PROLOG(0)
	li	r10,	0
	stw	r10,	MUTEX_ILK(r3)		/* clear interlock */
	stw	r10,	MUTEX_LOCKED(r3)	/* clear locked flag */
	sth	r10,	MUTEX_WAITERS(r3)	/* init waiter count */

#if	MACH_LDEBUG
	stw	r10,	MUTEX_PC(r3)		/* init caller pc */
	stw	r10,	MUTEX_THREAD(r3)	/* and owning thread */
	li	r10,	MUTEX_TAG
	stw	r10,	MUTEX_TYPE(r3)		/* set lock type */
#endif	/* MACH_LDEBUG */
#if	ETAP_LOCK_TRACE
	bl	EXT(etap_mutex_init)		/* init ETAP data */
#if 0
	lwz	r3,	12(r1)		/* restore r3 (save in prolog) */
#endif
#endif	/* ETAP_LOCK_TRACE */

	EPILOG
	blr

/*
 *	void _mutex_lock(mutex_t*)
 */

ENTRY(_mutex_lock,TAG_NO_FRAME_USED)

	PROLOG(12)
	
#if	ETAP_LOCK_TRACE
#define	SWT_HI	16(r1)
#define SWT_LO	20(r1)
#define MISSED	24(r1)
	li	r0,	0
	stw	r0,	SWT_HI		/* set wait time to 0 (HI) */
	stw	r0,	SWT_LO		/* set wait time to 0 (LO) */
	stw	r0,	MISSED		/* clear local miss marker */
#endif	/* ETAP_LOCK_TRACE */

	CHECK_MUTEX_TYPE()
	CHECK_NO_SIMPLELOCKS()

.L_ml_retry:
	DISABLE_PREEMPTION()
	li	r10,	1		/* value to be stored... 1==taken */
	li	r9,	MUTEX_ILK	/* in the interlock field */
.L_ml_get_hw:
	lwarx	r4,	r9,	r3	/* ld from addr of ARG0 and reserve */
	cmpwi	r4,	0		/* test ... */
	bne-	.L_ml_get_hw		/* loop if busy (predict not busy) */
	stwcx.	r10,	r9,	r3	/* and set (if reserved) */
	bne-	.L_ml_get_hw		/* if set failed, loop back */
	isync

/*
 * Beware of a race between this code path and the inline ASM fast-path locking
 * sequence which attempts to lock a mutex by directly setting the locked flag
 */
.L_ml_lock_retry:	
	li	r9,	MUTEX_LOCKED	/* the locked field */
	lwarx	r4,	r9,	r3	/* ld from addr of ARG0 and reserve */
	cmpwi	r4,	0		/* is the mutex locked ? */
	bne-	.L_ml_fail		/* yes, we loose */
	stwcx.	r10,	r9,	r3	/* lock it (if reserved) */
	bne-	.L_ml_lock_retry	/* if set failed, try again */
	isync

#if	MACH_LDEBUG
	mflr	r10
	stw	r10,	MUTEX_PC(r3)
	lwz	r10,	PP_CPU_DATA(r2)
	lwz	r10,	CPU_ACTIVE_THREAD(r10)
	stw	r10,	MUTEX_THREAD(r3)
	cmpwi	r10,	0
	beq-	.L_ml_no_active_thread
	lwz	r9,	THREAD_MUTEX_COUNT(r10)
	addi	r9,	r9,	1
	stw	r9,	THREAD_MUTEX_COUNT(r10)
.L_ml_no_active_thread:
#endif	/* MACH_LDEBUG */

	sync
	li	r10,	0
	stw	r10,	MUTEX_ILK(r3)	/* free the interlock */

	ENABLE_PREEMPTION()

#if	ETAP_LOCK_TRACE
	mflr	r4
	lwz	r5,	SWT_HI
	lwz	r6,	SWT_LO
	bl	EXT(etap_mutex_hold)	/* collect hold timestamp */
#if 0
	lwz	r3,	12(r1)		/* restore r3 (saved in prolog) */
#endif
#endif	/* ETAP_LOCK_TRACE */

	EPILOG
	blr

.L_ml_fail:
#if	ETAP_LOCK_TRACE
	lwz	r7,	MISSED
	cmpwi	r7,	0	/* did we already take a wait timestamp ? */
	bne	.L_ml_block	/* yup. carry-on */
	bl	EXT(etap_mutex_miss)	/* get wait timestamp */
	stw	r3,	SWT_HI	/* store timestamp */
	stw	r4,	SWT_LO
	li	r7,	1	/* mark wait timestamp as taken */
	stw	r7,	MISSED
	lwz	r3,	12(r1)	/* restore r3 (saved in prolog) */
#endif	/* ETAP_LOCK_TRACE */

.L_ml_block:
	CHECK_MYLOCK(MUTEX_THREAD)
	bl	EXT(mutex_lock_wait)	/* wait for the lock */
	lwz	r3,	12(r1)		/* restore r3 (saved in prolog) */
	
	b	.L_ml_retry		/* and try again */

	
/*
 *	void _mutex_try(mutex_t*)
 *
 * Do nothing for now
 */

ENTRY(_mutex_try,TAG_NO_FRAME_USED)

	PROLOG(8)	/* reserve space for SWT_HI and SWT_LO */
	
#if	ETAP_LOCK_TRACE
	li	r5,	0
	stw	r5,	STW_HI	/* set wait time to 0 (HI) */
	stw	r5,	SWT_LO	/* set wait time to 0 (LO) */
#endif	/* ETAP_LOCK_TRACE */

	CHECK_MUTEX_TYPE()
	CHECK_NO_SIMPLELOCKS()

	li	r10,	1
	li	r9,	MUTEX_LOCKED	/* the locked field */
	lwarx	r4,	r9,	r3	/* ld from addr of ARG0 and reserve */
	cmpwi	r4,	0		/* is the mutex locked ? */
	bne-	.L_mt_fail		/* yes, we loose */
	stwcx.	r10,	r9,	r3	/* lock it (if reserved) */
	bne-	.L_mt_fail		/* if set failed, we loose */
	isync

#if	MACH_LDEBUG
	mflr	r10
	stw	r10,	MUTEX_PC(r3)
	lwz	r10,	PP_CPU_DATA(r2)
	lwz	r10,	CPU_ACTIVE_THREAD(r10)
	stw	r10,	MUTEX_THREAD(r3)
	cmpwi	r10,	0
	beq-	.L_mt_no_active_thread
	lwz	r9,	THREAD_MUTEX_COUNT(r10)
	addi	r9,	r9,	1
	stw	r9,	THREAD_MUTEX_COUNT(r10)
.L_mt_no_active_thread:
#endif	/* MACH_LDEBUG */

#if	ETAP_LOCK_TRACE
	mflr	r4
	lwz	r5,	SWT_HI
	lwz	r6,	SWT_LO
	bl	EXT(etap_mutex_hold)	/* get start hold timestamp */
	lwz	r3,	12(r1)		/* restore r3 (saved in prolog) */
#endif	/* ETAP_LOCK_TRACE */

	li	r3,	1		/* success */
.L_mt_done:	
	EPILOG
	blr

.L_mt_fail:	
	li	r3,	0		/* failure */
	b	.L_mt_done
		
/*
 *	void mutex_unlock(mutex_t* l)
 */

ENTRY(mutex_unlock, TAG_NO_FRAME_USED)

	PROLOG(0)
	
#if	ETAP_LOCK_TRACE
	bl	EXT(etap_mutex_unlock)	/* collect ETAP data */
	lwz	r3,	12(r1)		/* restore r3 (saved in prolog) */
#endif	/* ETAP_LOCK_TRACE */

	CHECK_MUTEX_TYPE()
	CHECK_THREAD(MUTEX_THREAD)

	DISABLE_PREEMPTION()

	li	r10,	1		/* value to be stored... 1==taken */
	li	r9,	MUTEX_ILK	/* in the interlock field */
.L_mu_get_hw:
	lwarx	r4,	r9,	r3	/* ld from addr of ARG0 and reserve */
	cmpwi	r4,	0		/* test ... */
	bne-	.L_mu_get_hw		/* loop if busy (predict not busy) */
	stwcx.	r10,	r9,	r3	/* and set (if reserved) */
	bne-	.L_mu_get_hw		/* if set failed, loop back */
	isync

	lhz	r10,	MUTEX_WAITERS(r3)	/* are there any waiters ? */
	cmpwi	r10,	0
	bne-	.L_mu_wakeup			/* yes, more work to do */

.L_mu_doit:
#if	MACH_LDEBUG
	li	r10,	0
	stw	r10,	MUTEX_THREAD(r3)	/* disown thread */
	lwz	r10,	PP_CPU_DATA(r2)
	lwz	r10,	CPU_ACTIVE_THREAD(r10)
	cmpwi	r10,	0
	beq-	.L_mu_no_active_thread
	lwz	r9,	THREAD_MUTEX_COUNT(r10)
	subi	r9,	r9,	1
	stw	r9,	THREAD_MUTEX_COUNT(r10)
.L_mu_no_active_thread:
#endif	/* MACH_LDEBUG */

	sync
	li	r10,	0
	stw	r10,	MUTEX_LOCKED(r3)	/* unlock the mutex */

	stw	r10,	MUTEX_ILK(r3)		/* unlock the interlock */

	ENABLE_PREEMPTION()

	EPILOG
	blr

.L_mu_wakeup:
	bl	EXT(mutex_unlock_wakeup)	/* yes, wake a thread */
	lwz	r3,	12(r1)		/* restore r3 (saved in prolog) */
	b	.L_mu_doit

/*
 *	void interlock_unlock(hw_lock_t lock)
 */
ENTRY(interlock_unlock, TAG_NO_FRAME_USED)

	PROLOG(0)
	li	r10,	0
	stw	r10,	0(r3)

	ENABLE_PREEMPTION()

	EPILOG
	blr
