/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux
 */

/*
 * Low-memory exception vector code for PowerPC MACH
 *
 * These are the only routines that are ever run with
 * VM instruction translation switched off.
 *
 * The PowerPC is quite strange in that rather than having a set
 * of exception vectors, the exception handlers are installed
 * in well-known addresses in low memory. This code must be loaded
 * at ZERO in physical memory. The simplest way of doing this is
 * to load the kernel at zero, and specify this as the first file
 * on the linker command line.
 *
 * When this code is loaded into place, it is loaded at virtual
 * address KERNELBASE, which is mapped to zero (physical).
 *
 * This code handles all powerpc exceptions and is always entered
 * in supervisor mode with translation off. It saves the minimum
 * processor state before switching back on translation and
 * jumping to the approprate routine.
 *
 * Vectors from 0x100 to 0x3fff occupy 0x100 bytes each (64 instructions)
 *
 * We use some of this space to decide which stack to use, and where to
 * save the context etc, before	jumping to a generic handler.
 */

#include <assym.s>
#include <debug.h>
#include <cpus.h>
	
#include <mach_debug.h>
#include <ppc/asm.h>
#include <ppc/proc_reg.h>
#include <ppc/exception.h>
#include <mach/ppc/vm_param.h>
	
	.text

	.globl _ExceptionVectorsStart
	.type  _ExceptionVectorsStart,@function
_ExceptionVectorsStart:	/* Used if relocating the exception vectors */

/* 
 * System reset - call debugger
 */


	. = 0x100
.L_handler100:	

	mtsprg	3,	r3
	li	r3,	T_RESET
	b	.L_exception_entry

/*
 * Machine check (physical bus error) - call debugger
 */

	. = 0x200
.L_handler200:

	mtsprg	3,	r3
	li	r3,	T_MACHINE_CHECK
	b	.L_exception_entry

/*
 * Data access - page fault, invalid memory rights for operation
 */

	. = 0x300
.L_handler300:

	mtsprg	3,	r3
	li	r3,	T_DATA_ACCESS
	b	.L_exception_entry

/*
 * Instruction access - as for data access
 */

	. = 0x400
.L_handler400:

	mtsprg	3,	r3
	li	r3,	T_INSTRUCTION_ACCESS
	b	.L_exception_entry

/*
 * External interrupt
 */

	. = 0x500
.L_handler500:

	mtsprg	3,	r3
	li	r3,	T_INTERRUPT
	b	.L_exception_entry

/*
 * Alignment - many reasons
 */

	. = 0x600
.L_handler600:
	
	mtsprg	3,	r3
	li	r3,	T_ALIGNMENT
	b	.L_exception_entry

/*
 * Program - floating point exception, illegal inst, priv inst, user trap
 */

	. = 0x700
.L_handler700:

	mtsprg	3,	r3
	li	r3,	T_PROGRAM
	b	.L_exception_entry

/*
 * Program - floating point disabled, illegal inst, priv inst, user trap
 */

	. = 0x800
.L_handler800:

	mtsprg	3,	r3
	li	r3,	T_FP_UNAVAILABLE
	b	.L_exception_entry


/*
 * Decrementer - DEC register has passed zero.
 */

	. = 0x900
.L_handler900:

	mtsprg	3,	r3
	li	r3,	T_DECREMENTER
	b	.L_exception_entry

/*
 * I/O controller interface error - MACH does not use this
 */

	. = 0xA00
.L_handlerA00:

	mtsprg	3,	r3
	li	r3,	T_IO_ERROR
	b	.L_exception_entry

/*
 * Reserved
 */

	. = 0xb00
.L_handlerB00:

	mtsprg	3,	r3
	li	r3,	T_RESERVED
	b	.L_exception_entry

/*
 * System call - generated by the sc instruction
 */

	. = 0xc00
.L_handlerC00:

	mtsprg	3,	r3
	li	r3,	T_SYSTEM_CALL
	b	.L_exception_entry

/*
 * Trace - generated by single stepping a 603/604
 */

	. = 0xd00
.L_handlerD00:

	mtsprg	3,	r3
	li	r3,	T_TRACE
	b	.L_exception_entry

/*
 * Floating point assist
 */

	. = 0xe00
.L_handlerE00:

	mtsprg	3,	r3
	li	r3,	T_FP_ASSIST
	b	.L_exception_entry

/*
 * Instruction translation miss - we inline this code.
 * Upon entry (done for us by the machine):
 *     srr0 :	 addr of instruction that missed
 *     srr1 :	 bits 0-3   = saved CR0
 *                    4     = lru way bit
 *                    16-31 = saved msr
 *     msr[tgpr] = 1  (so gpr0-3 become our temporary variables)
 *     imiss:	 ea that missed
 *     icmp :	 the compare value for the va that missed
 *     hash1:	 pointer to first hash pteg
 *     hash2:	 pointer to 2nd hash pteg
 *
 * Register usage:
 *     tmp0:	 saved counter
 *     tmp1:	 junk
 *     tmp2:	 pointer to pteg
 *     tmp3:	 current compare value
 *
 * This code is taken from the 603e User's Manual with
 * some bugfixes and minor improvements to save bytes and cycles
 */

	. = 0x1000
.L_handler1000:
	mfspr	tmp2,	hash1
	mfctr	tmp0				/* use tmp0 to save ctr */
	mfspr	tmp3,	icmp

.L_imiss_find_pte_in_pteg:
	li	tmp1,	8			/* count */
	subi	tmp2,	tmp2,	8		/* offset for lwzu */
	mtctr	tmp1				/* count... */
	
.L_imiss_pteg_loop:
	lwzu	tmp1,	8(tmp2)			/* check pte0 for match... */
	cmpw	CR0,	tmp1,	tmp3
#if 0	
	bdnzf+	CR0,	.L_imiss_pteg_loop
#else	
	bc	CR0,2,	.L_imiss_pteg_loop
#endif	
	beq+	CR0,	.L_imiss_found_pte

	/* Not found in PTEG, we must scan 2nd then give up */

	andi.	tmp1,	tmp3,	MASK(PTE0_HASH_ID)
	bne-	.L_imiss_do_no_hash_exception		/* give up */

	mfspr	tmp2,	hash2
	ori	tmp3,	tmp3,	MASK(PTE0_HASH_ID)
	b	.L_imiss_find_pte_in_pteg

.L_imiss_found_pte:

	lwz	tmp1,	4(tmp2)				/* get pte1_t */
	andi.	tmp3,	tmp1,	MASK(PTE1_WIMG_GUARD)	/* Fault? */
	bne-	.L_imiss_do_prot_exception		/* Guarded - illegal */

	/* Ok, we've found what we need to, restore and rfi! */

	mtctr	tmp0					/* restore ctr */
	mfsrr1	tmp3
	mfspr	tmp0,	imiss
	mtcrf	0x80,	tmp3				/* Restore CR0 */
	mtspr	rpa,	tmp1				/* set the pte */
	ori	tmp1,	tmp1,	MASK(PTE1_REFERENCED)	/* set referenced */
	tlbli	tmp0
	sth	tmp1,	6(tmp2)
	rfi
	
.L_imiss_do_prot_exception:
	/* set up srr1 to indicate protection exception... */
	mfsrr1	tmp3
	andi.	tmp2,	tmp3,	0xffff
	addis	tmp2,	tmp2,	MASK(SRR1_TRANS_PROT) >> 16
	b	.L_imiss_do_exception
	
.L_imiss_do_no_hash_exception:
	/* clean up registers for protection exception... */
	mfsrr1	tmp3
	andi.	tmp2,	tmp3,	0xffff
	addis	tmp2,	tmp2,	MASK(SRR1_TRANS_HASH) >> 16
	
	/* And the entry into the usual instruction fault handler ... */
.L_imiss_do_exception:

	mtctr	tmp0					/* Restore ctr */
	mtsrr1	tmp2					/* Set up srr1 */
	mfmsr	tmp0					
	xoris	tmp0,	tmp0,	MASK(MSR_TGPR)>>16	/* no TGPR */
	mtcrf	0x80,	tmp3				/* Restore CR0 */
	mtmsr	tmp0					/* reset MSR[TGPR] */
	b	.L_handler400				/* Instr Access */
	
/*
 * Data load translation miss
 *
 * Upon entry (done for us by the machine):
 *     srr0 :	 addr of instruction that missed
 *     srr1 :	 bits 0-3   = saved CR0
 *                    4     = lru way bit
 *                    5     = 1 if store
 *                    16-31 = saved msr
 *     msr[tgpr] = 1  (so gpr0-3 become our temporary variables)
 *     dmiss:	 ea that missed
 *     dcmp :	 the compare value for the va that missed
 *     hash1:	 pointer to first hash pteg
 *     hash2:	 pointer to 2nd hash pteg
 *
 * Register usage:
 *     tmp0:	 saved counter
 *     tmp1:	 junk
 *     tmp2:	 pointer to pteg
 *     tmp3:	 current compare value
 *
 * This code is taken from the 603e User's Manual with
 * some bugfixes and minor improvements to save bytes and cycles
 */

	. = 0x1100
.L_handler1100:
	mfspr	tmp2,	hash1
	mfctr	tmp0				/* use tmp0 to save ctr */
	mfspr	tmp3,	dcmp

.L_dlmiss_find_pte_in_pteg:
	li	tmp1,	8			/* count */
	subi	tmp2,	tmp2,	8		/* offset for lwzu */
	mtctr	tmp1				/* count... */
	
.L_dlmiss_pteg_loop:
	lwzu	tmp1,	8(tmp2)			/* check pte0 for match... */
	cmpw	CR0,	tmp1,	tmp3
#if 0 /* How to write this correctly? */	
	bdnzf+	CR0,	.L_dlmiss_pteg_loop
#else	
	bc	CR0,2,	.L_dlmiss_pteg_loop
#endif	
	beq+	CR0,	.L_dmiss_found_pte

	/* Not found in PTEG, we must scan 2nd then give up */

	andi.	tmp1,	tmp3,	MASK(PTE0_HASH_ID)	/* already at 2nd? */
	bne-	.L_dmiss_do_no_hash_exception		/* give up */

	mfspr	tmp2,	hash2
	ori	tmp3,	tmp3,	MASK(PTE0_HASH_ID)
	b	.L_dlmiss_find_pte_in_pteg

.L_dmiss_found_pte:

	lwz	tmp1,	4(tmp2)				/* get pte1_t */

	/* Ok, we've found what we need to, restore and rfi! */

	mtctr	tmp0					/* restore ctr */
	mfsrr1	tmp3
	mfspr	tmp0,	dmiss
	mtcrf	0x80,	tmp3				/* Restore CR0 */
	mtspr	rpa,	tmp1				/* set the pte */
	ori	tmp1,	tmp1,	MASK(PTE1_REFERENCED)	/* set referenced */
	tlbld	tmp0					/* load up tlb */
	sth	tmp1,	6(tmp2)				/* sth is faster? */
	rfi
	
	/* This code is shared with data store translation miss */
	
.L_dmiss_do_no_hash_exception:
	/* clean up registers for protection exception... */
	mfsrr1	tmp3
	/* prepare to set DSISR_WRITE_BIT correctly from srr1 info */
	rlwinm	tmp1,	tmp3,	9,	6,	6
	addis	tmp1,	tmp1,	MASK(SRR1_TRANS_NO_PTE) >> 16

	/* And the entry into the usual data fault handler ... */

	mtctr	tmp0					/* Restore ctr */
	andi.	tmp2,	tmp3,	0xffff			/* Clean up srr1 */
	mtsrr1	tmp2					/* Set srr1 */
	mtdsisr	tmp1
	mfspr	tmp2,	dmiss
	mtdar	tmp2
	mfmsr	tmp0
	xoris	tmp0,	tmp0,	MASK(MSR_TGPR)>>16	/* no TGPR */
	mtcrf	0x80,	tmp3				/* Restore CR0 */
	sync						/* Needed on some */
	mtmsr	tmp0					/* reset MSR[TGPR] */
	b	.L_handler300				/* Data Access */
	
/*
 * Data store translation miss (similar to data load)
 *
 * Upon entry (done for us by the machine):
 *     srr0 :	 addr of instruction that missed
 *     srr1 :	 bits 0-3   = saved CR0
 *                    4     = lru way bit
 *                    5     = 1 if store
 *                    16-31 = saved msr
 *     msr[tgpr] = 1  (so gpr0-3 become our temporary variables)
 *     dmiss:	 ea that missed
 *     dcmp :	 the compare value for the va that missed
 *     hash1:	 pointer to first hash pteg
 *     hash2:	 pointer to 2nd hash pteg
 *
 * Register usage:
 *     tmp0:	 saved counter
 *     tmp1:	 junk
 *     tmp2:	 pointer to pteg
 *     tmp3:	 current compare value
 *
 * This code is taken from the 603e User's Manual with
 * some bugfixes and minor improvements to save bytes and cycles
 */

	. = 0x1200
.L_handler1200:
	mfspr	tmp2,	hash1
	mfctr	tmp0				/* use tmp0 to save ctr */
	mfspr	tmp3,	dcmp

.L_dsmiss_find_pte_in_pteg:
	li	tmp1,	8			/* count */
	subi	tmp2,	tmp2,	8		/* offset for lwzu */
	mtctr	tmp1				/* count... */
	
.L_dsmiss_pteg_loop:
	lwzu	tmp1,	8(tmp2)			/* check pte0 for match... */
	cmpw	CR0,	tmp1,	tmp3
#if 0 /* I don't know how to write this properly */	
	bdnzf+	CR0,	.L_dsmiss_pteg_loop
#else	
	bc	CR0,2,	.L_dsmiss_pteg_loop
#endif	
	beq+	CR0,	.L_dsmiss_found_pte

	/* Not found in PTEG, we must scan 2nd then give up */

	andi.	tmp1,	tmp3,	MASK(PTE0_HASH_ID)	/* already at 2nd? */
	bne-	.L_dmiss_do_no_hash_exception		/* give up */

	mfspr	tmp2,	hash2
	ori	tmp3,	tmp3,	MASK(PTE0_HASH_ID)
	b	.L_dsmiss_find_pte_in_pteg

.L_dsmiss_found_pte:

	lwz	tmp1,	4(tmp2)				/* get pte1_t */
	andi.	tmp3,	tmp1,	MASK(PTE1_CHANGED)	/* unchanged, check? */
	beq-	.L_dsmiss_check_prot			/* yes, check prot */

.L_dsmiss_resolved:
	/* Ok, we've found what we need to, restore and rfi! */

	mtctr	tmp0					/* restore ctr */
	mfsrr1	tmp3
	mfspr	tmp0,	dmiss
	mtcrf	0x80,	tmp3				/* Restore CR0 */
	mtspr	rpa,	tmp1				/* set the pte */
	tlbld	tmp0					/* load up tlb */
	rfi
	
.L_dsmiss_check_prot:
	/* PTE is unchanged, we must check that we can write */
	rlwinm.	tmp3,	tmp1,	30,	0,	1	/* check PP[1] */
	bge-	.L_dsmiss_check_prot_user_kern
	andi.	tmp3,	tmp1,	1			/* check PP[0] */
	beq+	.L_dsmiss_check_prot_ok
	
.L_dmiss_do_prot_exception:
	/* clean up registers for protection exception... */
	mfsrr1	tmp3
	/* prepare to set DSISR_WRITE_BIT correctly from srr1 info */
	rlwinm	tmp1,	tmp3,	9,	6,	6
	addis	tmp1,	tmp1,	MASK(SRR1_TRANS_NO_PTE) >> 16

	/* And the entry into the usual data fault handler ... */

	mtctr	tmp0					/* Restore ctr */
	andi.	tmp2,	tmp3,	0xffff			/* Clean up srr1 */
	mtsrr1	tmp2					/* Set srr1 */
	mtdsisr	tmp1
	mfspr	tmp2,	dmiss
	mtdar	tmp2
	mfmsr	tmp0
	xoris	tmp0,	tmp0,	MASK(MSR_TGPR)>>16	/* no TGPR */
	mtcrf	0x80,	tmp3				/* Restore CR0 */
	sync						/* Needed on some */
	mtmsr	tmp0					/* reset MSR[TGPR] */
	b	.L_handler300				/* Data Access */
	
/* NB - if we knew we were on a 603e we could test just the MSR_KEY bit */
.L_dsmiss_check_prot_user_kern:
	mfsrr1	tmp3
	andi.	tmp3,	tmp3,	MASK(MSR_PR)
	beq+	.L_dsmiss_check_prot_kern
	mfspr	tmp3,	dmiss				/* check user privs */
	mfsrin	tmp3,	tmp3				/* get excepting SR */
	andis.	tmp3,	tmp3,	0x2000			/* Test SR ku bit */
	beq+	.L_dsmiss_check_prot_ok
	b	.L_dmiss_do_prot_exception

.L_dsmiss_check_prot_kern:
	mfspr	tmp3,	dmiss				/* check kern privs */
	mfsrin	tmp3,	tmp3
	andis.	tmp3,	tmp3,	0x4000			/* Test SR Ks bit */
	bne-	.L_dmiss_do_prot_exception

.L_dsmiss_check_prot_ok:
	/* Ok, mark as referenced and changed before resolving the fault */
	ori	tmp1,	tmp1,	(MASK(PTE1_REFERENCED)|MASK(PTE1_CHANGED))
	sth	tmp1,	6(tmp2)
	b	.L_dsmiss_resolved
	
/*
 * Instruction address breakpoint
 */

	. = 0x1300
.L_handler1300:
	mtsprg	3,	r3
	li	r3,	T_INSTRUCTION_BKPT
	b	.L_exception_entry


/*
 * System management interrupt
 */

	. = 0x1400
.L_handler1400:
	mtsprg	3,	r3
	li	r3,	T_SYSTEM_MANAGEMENT
	b	.L_exception_entry


/*
 * There is now a large gap of reserved traps
 */

/*
 * Run mode/ trace exception - single stepping on 601 processors
 */

	. = 0x2000
.L_handler2000:

	mtsprg	3,	r3
	li	r3,	T_RUNMODE_TRACE
	b	.L_exception_entry

/*
 * .L_exception_entry(type)
 *
 * This is the common exception handling routine called by any
 * type of system exception.
 *
 * ENTRY:	via a system exception handler, thus interrupts off, VM off.
 *              r3 has been saved in sprg3 and now contains a number
 *              representing the exception's origins
 *
 * EXIT:	srr0 and srr1 saved in per_proc_info structure
 *              r3 (supplied) saved in per_proc_info structure
 *              cr            saved in per_proc_info structure
 *              original r1-3 saved in sprg1-3.
 *                 r1 - is scratch
 *                 r2 - byte offset into per_proc_info of our CPU (SMP only)
 *                 r3  -contains exception info as for entry
 *
 *              The exception's handler is entered with
 *              VM on, interrupts still switched off
 */

#define PRINT1H(XX, nn)		\
0:	lbz	r2,0(r1)	; \
	andi.	r2,r2,0x04	; \
	beq	00b		; \
	XX			; \
	srwi	r2,r2,nn	; \
	andi.	r2,r2,0x0F	; \
	cmpi	0,r2,0x0A	; \
	blt	1f		; \
	subi	r2,r2,0x0A	; \
	addi	r2,r2,'A'-'0'	; \
1:	addi	r2,r2,'0'	; \
	stb	r2,4(r1)	;
#define PRINT1H_(XX)		\
0:	lbz	r2,0(r1)	; \
	andi.	r2,r2,0x04	; \
	beq	00b		; \
	XX			; \
	andi.	r2,r2,0x0F	; \
	cmpi	0,r2,0x0A	; \
	blt	1f		; \
	subi	r2,r2,0x0A	; \
	addi	r2,r2,'A'-'0'	; \
1:	addi	r2,r2,'0'	; \
	stb	r2,4(r1)	;
#define PRINT8H(XX)		\
	PRINT1H(XX, 28)		\
	PRINT1H(XX, 24)		\
	PRINT1H(XX, 20)		\
	PRINT1H(XX, 16)		\
	PRINT1H(XX, 12)		\
	PRINT1H(XX,  8)		\
	PRINT1H(XX,  4)		\
	PRINT1H_(XX)
#define PRINT1C(XX)		\
0:	lbz	r2,0(r1)	; \
	andi.	r2,r2,0x04	; \
	beq	00b		; \
	li	r2,XX		; \
	stb	r2,4(r1)	;
	
	.text
		
.L_exception_entry:
	/* r3 has been saved in sprg3 and now contains exception info */
	mtsprg	1,	r1
	mtsprg	2,	r2

	/* Save SRR0 and SRR1 plus cr and r3 into PER_PROC structure */
	
	mfsprg	r2,	0		/* sprg0 = phys addr of per-proc */
	stw	r3,	PP_SAVE_EXCEPTION_TYPE(r2)
	mfsrr0	r1
	stw	r1,	PP_SAVE_SRR0(r2)
	mfsrr1	r1
	stw	r1,	PP_SAVE_SRR1(r2)
	mfcr	r1
	stw	r1,	PP_SAVE_CR(r2)

#if 0	
/* TEMP */
/* Turn off DATA cache */	
	mfspr	r1,1008
	li	r2,0x4000
	andc	r1,r1,r2
	mtspr	1008,r1
	lis	r1,0x50F0C000>>16
	ori	r1,r1,0x50F0C000&0xFFFF
/* Print exception vector # */
#define AA	mr r2,r3	
	PRINT1H(AA, 4)
	PRINT1H_(AA)
/* Print SRR0 */
#undef AA
#define AA	mfsrr0 r2
	PRINT1C(' ')
#if 0 /* I couldn't see how to make this work */	
	PRINT8H(AA)
#else
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
#endif		
/* Print SPRG0 */	
#undef AA
#define AA	mfsprg r2,0
	PRINT1C(' ')
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
/* Print SPRG2 */	
#undef AA
#define AA	mfsprg r2,2
	PRINT1C(' ')
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
/* Print SPRG3 */	
#undef AA
#define AA	mfsprg r2,3
	PRINT1C(' ')
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
/* Print CR/LF */	
	PRINT1C(0x0D)
	PRINT1C(0x0A)
/* Dump PP_DATA */
	PRINT1C(' ')
#undef AA	
#define AA	\
	mfsprg	r2,0; lwz r2,16(r2)
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
	PRINT1C(' ')
#undef AA	
#define AA	\
	mfsprg	r2,0; lwz r2,20(r2)
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
	PRINT1C(' ')
#undef AA	
#define AA	\
	mfsprg	r2,0; lwz r2,24(r2)
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
	PRINT1C(' ')
#undef AA	
#define AA	\
	mfsprg	r2,0; lwz r2,28(r2)
	PRINT1H(AA, 28)
	PRINT1H(AA, 24)
	PRINT1H(AA, 20)
	PRINT1H(AA, 16)
	PRINT1H(AA, 12)
	PRINT1H(AA,  8)
	PRINT1H(AA,  4)
	PRINT1H_(AA)
/* Print CR/LF */	
	PRINT1C(0x0D)
	PRINT1C(0x0A)
/* Turn caches back on */	
	mfspr	r1,1008
	ori	r1,r1,0x4000
	mtspr	1008,r1
	mfsprg	r2,	0		/* sprg0 = phys addr of per-proc */
/* TEMP */	
#endif
		
	/* Remap the kernel using seg reg 0, and I/O via seg reg 5 */

	lis	r1,	(KERNEL_SEG_REG0_VALUE >> 16)
#if PPC_SID_KERNEL != 0
		/* if PPC_SID_KERNEL == 0, then bottom 16 bits are 0,
		 * so save an instruction - big deal
		 */
	ori	r1,	r1,	(KERNEL_SEG_REG0_VALUE & 0xFFFF)
#endif
	mtsr	0,	r1		/* Kernel SR0 */
	ori	r1,	r1,	1	/* Kernel SR1 */
	mtsr	1,	r1
		
	/* jump into main handler code switching on VM at the same time */

	/* We assume kernel data is mapped contiguously in physical
	 * memory, otherwise we'd need to switch on (at least) virtual data.
	 */
	lwz	r1,	PP_PHYS_EXCEPTION_HANDLERS(r2)
	lwzx	r1,	r1,	r3

	mtsrr0	r1
	
	lwz	r2,	PP_VIRT_PER_PROC(r2)
	
	li	r1,	MSR_SUPERVISOR_INT_OFF
	mtsrr1	r1
	rfi		/* into the exception handler with VM on */

/*
 * exception_exit(sr0,srr0,srr1)
 *
 * This is the trampoline code used when exiting into a foreign
 * address space.
 *
 * NMGS TODO - can't we skip exception_exit by relying on translations
 * NMGS TODO   even after we've loaded sr0? Docs aren't clear. Using
 * NMGS TODO   1-1 kernel text mapping would definately avoid this need.
 *
 * ENTRY :	 entry via rfi, MSR = MSR_VM_OFF
 *               r1-3 saved in sprg1-3
 * 		 r1 = user's sr0 - used to construct sr1 too
 *               r2 = user's srr0 (instruction pointer)
 *               r3 = user's srr1 (msr)
 *
 * EXIT :	 this routine restores the users' space and rfis.
 */
	
	.data
	.align	ALIGN
	.type	EXT(exception_exit),@object
	.size	EXT(exception_exit), 4
	.globl	EXT(exception_exit)
EXT(exception_exit):
	.long	exception_exit_fn-KERNELBASE_TEXT	/* phys addr of fn */


	.text	
	
exception_exit_fn:

	mtsrr0	r2
	mtsrr1	r3

	mtsr	0,	r1		/* Restore user space SR0 */
	addi	r1,	r1,	1
	mtsr	1,	r1		/* Restore user space SR1 */
		
	mfsprg	r1,	1
	mfsprg	r2,	2
	mfsprg	r3,	3
	rfi
	
	.globl _ExceptionVectorsEnd
	.type  _ExceptionVectorsEnd,@function
_ExceptionVectorsEnd:	/* Used if relocating the exception vectors */


