/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux
 */

#include <cpus.h>
#include <mach_kdb.h>
#include <mach_kgdb.h>

#include <ppc/asm.h>
#include <ppc/proc_reg.h>
#include <mach/ppc/vm_param.h>
#include <assym.s>
	
/*
 * Interrupt and bootup stack for initial processor
 */

	.file	"start.s"
	
	.section ".data"
		/* Align on page boundry */
	.align  PPC_PGSHIFT
		/* Red zone for interrupt stack, one page (will be unmapped)*/
	.set	., .+PPC_PGBYTES
		/* intstack itself */
	.type 	EXT(intstack),@object
	.size	EXT(intstack),INTSTACK_SIZE
        .globl  EXT(intstack)
EXT(intstack):
	.set	., .+INTSTACK_SIZE
	.align	ALIGN
	.type 	EXT(intstack_top_ss),@object
	.size	EXT(intstack_top_ss),4
        .globl  EXT(intstack_top_ss)
	/* intstack_top_ss points to the topmost saved state structure
	 * in the intstack */
EXT(intstack_top_ss):
	.long	intstack+INTSTACK_SIZE-SS_SIZE
	
	/* KGDB stack - used by the debugger if present */
#if MACH_KGDB
	.type 	EXT(gdbstack),@object
	.size	EXT(gdbstack),KERNEL_STACK_SIZE
        .globl  EXT(gdbstack)
EXT(gdbstack):
	.set	., .+KERNEL_STACK_SIZE
	.align	ALIGN
	.type 	EXT(gdbstack_top_ss),@object
	.size	EXT(gdbstack_top_ss),4
        .globl  EXT(gdbstack_top_ss)
	/* gdbstack_top_ss points to the topmost saved state structure
	 * in the gdbstack */
EXT(gdbstack_top_ss):
	.long	gdbstack+KERNEL_STACK_SIZE-SS_SIZE
	.type 	EXT(gdbstackptr),@object
	.size	EXT(gdbstackptr),4
        .globl  EXT(gdbstackptr)
EXT(gdbstackptr):
	.long	gdbstack+KERNEL_STACK_SIZE-SS_SIZE
#endif /* MACH_KGDB */

#if	MACH_KDB
/*
 * Kernel debugger stack for each processor.
 */
	.type	EXT(db_stack_store),@object
	.size	EXT(db_stack_store),INTSTACK_SIZE*NCPUS
	.globl	EXT(db_stack_store)
EXT(db_stack_store):
	.set	.,.+(INTSTACK_SIZE*NCPUS)
	.align	ALIGN
#endif	/* MACH_KDB */

/*
 * All CPUs start here.
 *
 * This code is called from the MachOS Loader, with a certain amount
 * of machine state already set up:
 *
 *    Address translation is switched ON, with:
 *        BAT0 mapping Physical 0-1M to 8-9M virtual (KERNELBASE_TEXT_OFFSET)
 *        BAT1 mapping Physical 4-5M to 4-5M virtual (KERNELBASE_DATA_OFFSET)
 *	
 *         and I/O correctly memory mapped via segs (so printf works already!)
 *
 *    We're running on a temporary stack.
 *
 *    Various arguments are passed via a table:
 *          ARG0 = contiguous physical memory size from 0
 *          ARG1 = pointer to other startup parameters
 */
	.text
	
ENTRY(_start,TAG_NO_FRAME_USED)


	/* Make sure our MSR is valid */
	li	r0,	MSR_SUPERVISOR_INT_OFF
	sync
	mtmsr	r0
	isync
	
#if 0
	/*
	 * We're coming in from open firmware with VM on!
	 * Map the bottom 256M and the top 256M as 1-1
	 * using BATs so that we can go ahead and mess
	 * around with the page tables
	 */
	
#if 0
	lis	r3,	((0 | (0x7ff << (31-19)))>>16) & 0xffff
	ori	r3,	r3, (0 | (0x7ff << (31-19)))&0xffff
	li	r4,	0
	mtibatu 0,	r3
	mtibatl 0,	r4
	mtdbatu 0,	r3
	mtdbatl 0,	r4
	isync
#endif 
	lis	r3,	((0xf0000000 | (0x7ff << (31-19)))>>16) & 0xffff
	ori	r3,	r3, ((0xf0000000 | (0x7ff << (31-19))))&0xffff
	lis	r4,	0xf000
	ori	r4,	r3, 0x20
		
	mtibatu 2,	r3
	mtibatl 2,	r4
	mtdbatu 2,	r3
	mtdbatl 2,	r4
	isync

#endif	
#if 0
        lis     r31,    0xf301
        ori     r31,    r31,0x2004
0:	
	li	r0,	64
	stb	r0,	0(r31)
	eieio

	li	r0,	65
	stb	r0,	0(r31)
	eieio
	li	r0,	64
	stb	r0,	2(r31)
	eieio

	li	r0,	65
	stb	r0,	2(r31)
	eieio

#endif /* 0 */	
	/* move onto interrupt stack to release stack in high mem */

	/* Load address of first thing in the various BSSs (not just bss)
	 * We assume that this starts right after the data ends
	 */
	lis	r10,	_edata@ha
	addi	r10,	r10,	_edata@l

	/* load  end of data */
	lis	r9,	_end@ha
	addi	r9, r9,	_end@l

	/* Get total byte count */
	subf	r9,	r10,	r9

	/* Convert byte count into words */
	rlwinm.	r9,	r9,30,2,31
	li	r0,0
	mtctr	r9
	addi	r10,r10,-4

1:
	stwu	r0,4(r10)
	bdnz	1b

	/* initialise */

	mr	r31,	ARG0		/* Save away arguments */
	mr	r30,	ARG1

	addis	r29,	0,	intstack_top_ss@ha
	addi	r29,	r29,	intstack_top_ss@l
	lwz	r29,	0(r29)

	li	r28,	0
	stw	r28,	FM_BACKPTR(r29) /* store a null frame backpointer */

	/* move onto new stack */
	
	mr	r1,	r29

	bl	EXT(go)

	mr	ARG0,	r31		/* Recover arguments */
	mr	ARG1,	r30

	/* Pass through ARG0, ARG1 */
	
	bl	EXT(machine_startup)

	/* Should never return */

	BREAKPOINT_TRAP
